
#include "CFlatDimstack.H"
#include "ogl.H"
#include "XmdvTool.H"
#include "brush.H"
#include "brush_oper.H"
#include "multi_select.H"
#include <string.h>

CFlatDimstack::CFlatDimstack(CFlatDisplayInformation* flat_info) : CFlatDisplay(flat_info)
{
}

CFlatDimstack::~CFlatDimstack()
{
}

//
// Determine if this value (val) for this dimension (dim) is covered by
// the brush.
//
static bool is_within(int val, int dim, int brush)
{
    double fval = val;
    double range = dataset->dim_max[dim] - dataset->dim_min[dim];
    double b_pos = dataset->cardinality[dim] * (brushes[brush].pos[dim]-dataset->dim_min[dim]) / range;
    double b_size = .5 * dataset->cardinality[dim] * brushes[brush].size[dim] / range;
    double start = b_pos - b_size;
    double end = b_pos + b_size;
    if (end > fval && end <= (fval+1.0)) return true;
    if (start >= fval && start < (fval+1.0)) return true;
    if (start < fval && end > (fval+1.0)) return true;
    return false;
}

//
// Recursive check for brush coverage, 2 dimensions at a time
//
void CFlatDimstack::rec_fill(grCanvas *canvas, double basex, double basey, int nextx, int nexty,
          ARR<double> &size, int brush)
{
    // If we had an even number of dimensions and are at the bottom of the
    // recursion, draw a rectangle in that data spot, as it is covered.
    //
    if (nextx == dataset->dims)
    {
        canvas->FillRectangle(basex, screen_distortion->dimstack_Y_size - basey - screen_distortion->dimstack_Y_blocksize, 
							  screen_distortion->dimstack_X_blocksize, screen_distortion->dimstack_Y_blocksize);
        return;
    }

    // If we had an odd number of dimensions and are at the bottom of the
    // recursion, check each descrete value for brush overlap and draw rectangles
    // over covered spots.
    //
    int i;
    if (nexty == dataset->dims)
    {
        for (i=0; i<dataset->cardinality[nextx]; i++) {
            if (is_within(i, nextx, brush) == true)
                canvas->FillRectangle(basex + i * size[nextx],
                    screen_distortion->dimstack_Y_size - basey - screen_distortion->dimstack_Y_blocksize, screen_distortion->dimstack_X_blocksize, 
					screen_distortion->dimstack_Y_blocksize);
        }
        return;
    }

    // Continue recursion for each bucket of the current 2 dimensions
    //
    for (i=0; i<dataset->cardinality[nextx]; i++)
        for (int j=0; j<dataset->cardinality[nexty]; j++)
            if ((is_within(i, nextx, brush) == 1) && (is_within(j, nexty, brush) == 1))
                rec_fill(canvas, basex + i * size[nextx], basey + j * size[nexty],
                    nextx+2, nexty+2, size, brush);
}

//
// Display brush coverage by descretizing brush in each dimension and check
// where in range of that dimension it overlaps.  Roundoff error in the
// discretization process can reduce the visual effect of brush coverage.
//
void CFlatDimstack::DrawSingleBrush(Brush *brush)
{
    canvas->SetForeground( interface_color[BRUSH1+brush->num]);
    // For the outermost dimensions...
    for (int i=0; i<dataset->cardinality[0]; i++) 
    {
        for (int j=0; j<dataset->cardinality[1]; j++)
        {
            // if the brush intersects those dimensions recursively
            // call to fill
            if ((is_within(i, 0, brush->num) == 1) && (is_within(j, 1, brush->num) == 1))
                rec_fill(canvas, i * screen_distortion->dimstack_size[0], j * screen_distortion->dimstack_size[1], 
						 2, 3, screen_distortion->dimstack_size, brush->num);
        }
    }
}

//
// Compute the descretized value from an original data value for a particular
// dimension.  Display will be messed up if it is out of bounds.  You then
// should fix the range for that dimension in your data file.
//
int get_contrib(int ind, double data)
{
    int res = int( ((data - dataset->dim_min[ind]) / (dataset->dim_max[ind] - dataset->dim_min[ind])) *
              (double)(dataset->cardinality[ind]));
/*
    if (res >= dataset->cardinality[ind] || res < 0) {
        printf("Data point out of bounds: dimension %d: [%lf, %lf]: %lf\n", ind,
            dataset->dim_min[ind], dataset->dim_max[ind], data);
    }
*/
    return res;
}

void CFlatDimstack::DrawSingleDimstackBin(grCanvas *canvas, const ARR<double> &data, const ARR<double> &size)
{
    ARR<int> contrib(dataset->dims);
    int j;
    for (j=0; j<dataset->dims; j++)
        contrib[j] = get_contrib(j, data[j]);

    double endx = 0, endy = 0;
    for (j=0; j<dataset->dims; j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < dataset->dims)
            endy += contrib[j] * size[j];
    }
    canvas->FillRectangle(endx, screen_distortion->dimstack_Y_size - endy - screen_distortion->dimstack_Y_blocksize, 
						  screen_distortion->dimstack_X_blocksize, screen_distortion->dimstack_Y_blocksize);
}

void CFlatDimstack::DrawSingleData(ARR<double> data, int datano)
{
    DrawSingleDimstackBin(canvas, data, screen_distortion->dimstack_size);
}



void CFlatDimstack::Draw()
{
    if (!dataset) return;

    if (dataset->dims<2) {
        return;
    }

    CFlatDisplay::Draw();
}

void CFlatDimstack::SetParameters()
{
    CFlatDisplay::SetParameters();
}

void CFlatDimstack::DrawAxes()
{
    DrawDimstackAxes(canvas);
}

void CFlatDimstack::DrawDimstackAxes(grCanvas *canvas)
{
    glLineWidth(1.);
    glShadeModel(GL_SMOOTH);
    // Compute the size of the blocks needed for the slowest dimension
    //
    double x_size = screen_distortion->dimstack_X_size;
    double y_size = screen_distortion->dimstack_Y_size;

    // Draw up to 3 levels of grid marks, based on the number of dimensions
    if (dataset->dims >= 6)
    {
        canvas->SetForeground(interface_color[GRID3]);
        for (double x = 0;x <= x_size; x = x + screen_distortion->dimstack_size[4])
            canvas->DrawLine(x, 0, x, y_size);

        for (double y = 0;y <= y_size; y = y + screen_distortion->dimstack_size[5])
            canvas->DrawLine(0, y, x_size, y);
    }

    if (dataset->dims >= 4)
    {
        canvas->SetForeground(interface_color[GRID2]);
        for (double x = 0;x <= x_size; x = x + screen_distortion->dimstack_size[2])
            canvas->DrawLine(x, 0, x, y_size);

        for (double y = 0;y <= y_size; y = y + screen_distortion->dimstack_size[3])
            canvas->DrawLine(0, y, x_size, y);
    }

    canvas->SetForeground(interface_color[GRID1]);

    for (double x = 0;x <= x_size; x = x + screen_distortion->dimstack_size[0])
        canvas->DrawLine(x, 0, x, y_size);

    for (double y = 0;y <= y_size; y = y + screen_distortion->dimstack_size[1])
        canvas->DrawLine(0, y, x_size, y);

}


static bool DimstackMapScreenToData(double x,double y, ARR<double> &center)
{
    // Now find the bin that the user clicked in
    int i;
    for (i=0; i<dataset->dims; i++)
    {
        // If odd, it's in the y direction
        if (i%2)
        {
            int bin = int(y/screen_distortion->dimstack_size[i]);

            // invert y
            int orig_bin = bin;
            bin = dataset->cardinality[i] - 1 - bin;

            // check bounds
            if (bin < 0 || bin >= dataset->cardinality[i])
                return false;

            // Increment data coords based on this
            center[i] = bin * (dataset->dim_max[i] - dataset->dim_min[i])/dataset->cardinality[i];
            center[i] += dataset->dim_min[i];

            // use original bin # for y decrement
            y -= orig_bin*screen_distortion->dimstack_size[i];
        }

        // If even, it's in the x direction
        else
        {
            int bin = int(x/screen_distortion->dimstack_size[i]);

            // check bounds
            if (bin < 0 || bin >= dataset->cardinality[i])
                return false;

            // Increment data coords based on this
            center[i] = bin * (dataset->dim_max[i] - dataset->dim_min[i])/dataset->cardinality[i];
            center[i] += dataset->dim_min[i];

            x -= bin*screen_distortion->dimstack_size[i];
        }
    }
    return true;
}

static void InputDimstackBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (curr_brush->display == false || curr_brush->enable == false)
        return;

    // Now find the bin that the user clicked in
    ARR<double> new_center(dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,new_center) )
        return; 

    // Recenter the brush
    int i;
    for (i=0; i<dataset->dims; i++)
        curr_brush->pos[i] = new_center[i];

    // Check if any of the brush values went out of range
    for (i=0; i<dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (curr_brush->Max(i) > dataset->dim_max[i])
            curr_brush->pos[i] = dataset->dim_max[i] - curr_brush->size[i]/2.;
        if (curr_brush->Min(i) < dataset->dim_min[i])
            curr_brush->pos[i] = dataset->dim_min[i] + curr_brush->size[i]/2.;
    }
}

static void InputDimstackMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (!multi_select) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Now find the bin that the user clicked in
    ARR<double> center(dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,center) )
        return;

/*
    Vec2 screen_rad = canvas->MapScreenToData( Vec2(multisel_radius,multisel_radius) ) -
                      canvas->MapScreenToData( Vec2(0,0) );
*/
    const double radN = .8;

    ARR<double> data(dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        double dist = 0;
        for (int j=0;j<dataset->dims;j++) {
            double d = (center[j] - data[j]);
            dist += d*d;
        }
        dist = sqrt(dist);

        if (dist < radN)
        {
            multiSelectAddPoint(i);

            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
			CFlatDimstack::DrawSingleDimstackBin(canvas,data,screen_distortion->dimstack_size);
            glDrawBuffer(GL_BACK);
        }
    }
}

static void DimstackCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Now find the bin that the user clicked in
    ARR<double> center(dataset->dims);
    if ( ! DimstackMapScreenToData(p.X,p.Y,center) )
        return;;

    char caption[50000]="Bin center: (";

    for (int i=0; i<dataset->dims; i++) {
        char str[100];
        sprintf(str, "%s: %.2lf ", dataset->names[i], center[i]);
        strcat(caption, str);
    }
    strcat(caption,")");

    appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),caption, canvas->name);
}

void InputDimstackButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    if (canvas->shiftkey && button == 0) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (button==2) {
        canvas->SetHandCursor();
        return;
    } else

    if (button == 0) {
        InputDimstackBrushSelect(canvas,x,y);
        BrushOperMarkDataset();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
    }
}

void CFlatDimstack::InputDimstackMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    DimstackCaption(canvas, x, y);

    if (canvas->shiftkey && canvas->button[0] == true) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (canvas->button[2] == true) {
        appCanvasMousePan(canvas);
    }
}

void InputDimstackButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (canvas->shiftkey && button == 0) {
        return;
    }
    canvas->SetDefaultCursor();
}


void CFlatDimstack::InputMouseMotion(int x, int y)
{
    InputDimstackMouseMotion(canvas,x,y);
}


void CFlatDimstack::InputButtonPress(int button, int x, int y)
{
    InputDimstackButtonPress(canvas, button, x,y);
}


void CFlatDimstack::InputButtonRelease(int button, int x, int y)
{
    InputDimstackButtonRelease(canvas, button, x, y);
}
